'''
Dependency of PINGMapper: https://github.com/CameronBodine/PINGMapper

Repository: https://github.com/CameronBodine/PINGVerter
PyPi: https://pypi.org/project/pingverter/ 

Developed by Cameron S. Bodine

###############
Acknowledgments
###############

None of this work would have been possible without the following repositories:

PyHum: https://github.com/dbuscombe-usgs/PyHum
SL3Reader: https://github.com/halmaia/SL3Reader
sonarlight: https://github.com/KennethTM/sonarlight


MIT License

Copyright (c) 2024 Cameron S. Bodine

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os, sys
import numpy as np
import pandas as pd
from datetime import datetime, timezone, timedelta
import pyproj
from io import BytesIO

# Add 'pingmapper' to the path, may not need after pypi package...
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(SCRIPT_DIR)
sys.path.append(PACKAGE_DIR)

from pingverter.verter_utils import filterGPS

# # RSD structur
# rsdStruct = np.dtype([
#     ("test", "<u4"),
# ])

garCols2PM = {
    'bottom_depth': 'inst_dep_m',
    'drawn_bottom_depth': 'keel_depth_m',
    'sample_cnt': 'ping_cnt', 
    'first_sample_depth': 'min_range',
    'last_sample_depth': 'max_range',
    'water_temp': 'tempC',
    'recording_time_ms': 'time_s',
}

class gar(object):

    #===========================================================================
    def __init__(self, inFile: str, nchunk: int=0, exportUnknown: bool=False):
        
        '''
        '''

        self.humFile = None
        self.sonFile = inFile
        self.nchunk = nchunk
        self.exportUnknown = exportUnknown

        self.magicNum = 3085556358



        self.extension = os.path.basename(inFile).split('.')[-1]

        # self.son_struct = rsdStruct

        self.garCols2PM = garCols2PM

        self.humDat = {} # Store general sonar recording metadata

        self.son8bit = False

        # Set Sonar beams
        # Neils beams: 1,3,4,5
        # UD beams: 0,2,4,6
        # Unsure and need to test
        # self.beam_set = {
        #     2: ['ds_lowfreq', 0],
        #     1: ['ds_hifreq', 1],
        #     5: ['ss_port', 2],
        #     6: ['ss_port', 2],
        #     4: ['ss_star', 3],
        #     0: ['ds_vhighfreq', 4],
        #     3: ['ds_vhighfreq', 5],
        # }

        return
    
    # ======================================================================
    def _getFileLen(self):
        self.file_len = os.path.getsize(self.sonFile)

        return
    
    # ======================================================================
    def _fread_dat(self,
            infile,
            num,
            typ):
        '''
        Helper function that reads binary data in a file.

        ----------------------------
        Required Pre-processing step
        ----------------------------
        Called from self._getHumDat(), self._cntHead(), self._decodeHeadStruct(),
        self._getSonMeta(), self._loadSonChunk()

        ----------
        Parameters
        ----------
        infile : file
            DESCRIPTION - A binary file opened in read mode at a pre-specified
                            location.
        num : int
            DESCRIPTION - Number of bytes to read.
        typ : type
            DESCRIPTION - Byte type

        -------
        Returns
        -------
        List of decoded binary data

        --------------------
        Next Processing Step
        --------------------
        Returns list to function it was called from.
        '''

        buffer = infile.read(num)
        data = np.frombuffer(buffer, dtype=typ)

        return data
    
    
    ### File Header ###
    # ======================================================================
    def _parseFileHeader(self):
        '''
        '''

        self.headBytes = 20480 # Hopefully a fixed value for all RSD files
        chanInfoLen = 1069 # It is not clear if there is helpful info in channel information...

        # Get the file header structure
        headStruct, firstHeadBytes = self._getFileHeaderStruct()

        # print('\n\n\nheadStruct:')
        # for v in headStruct:
        #     print(v)
        
        # Read the header
        file = open(self.sonFile, 'rb') # Open the file
        file.seek(0)

        # Get the data
        buffer = file.read(firstHeadBytes)

        # Read the data
        header = np.frombuffer(buffer=buffer, dtype=headStruct)

        out_dict = {}
        for name, typ in header.dtype.fields.items():
            out_dict[name] = header[name][0].item()

        # for k,v in out_dict.items():
        #     print(k, v)

        self.file_header = out_dict

        # Parse channel_information_array (field 6) for per-channel metadata.
        self.channel_info = self._parseChannelInformation(firstHeadBytes)


        return

    # ======================================================================
    def _read_varuint(self, file_obj):
        """Read protobuf-style VarUInt32/VarUInt64 from a file-like object."""
        result = 0
        shift = 0

        while True:
            b = file_obj.read(1)
            if not b:
                raise EOFError('Unexpected EOF while reading varuint.')

            byte = b[0]
            result |= (byte & 0x7F) << shift

            if not (byte & 0x80):
                return result

            shift += 7
            if shift > 63:
                raise ValueError('Invalid varuint: too many continuation bytes.')

    # ======================================================================
    def _decode_varuint_bytes(self, data: bytes):
        stream = BytesIO(data)
        return self._read_varuint(stream)

    # ======================================================================
    def _decode_varint32_bytes(self, data: bytes):
        """Decode zig-zag encoded VarInt32 from bytes."""
        u = self._decode_varuint_bytes(data)
        return (u >> 1) ^ -(u & 1)

    # ======================================================================
    def _decode_depth_varint(self, data: bytes):
        """Decode Garmin depth-like varints.

        Some RSD files appear to store these values as unsigned varints while
        others align with zig-zag VarInt32 semantics from the reverse-engineered
        spec. If zig-zag yields a negative depth, fall back to unsigned.
        """
        u = self._decode_varuint_bytes(data)
        z = self._decode_varint32_bytes(data)

        candidates = [v for v in (u, z) if v >= 0]
        if not candidates:
            return u

        # Keep depth-like values within a broad physical bound (<= 5 km in mm)
        # when possible, otherwise pick the smaller non-negative decoding.
        plausible = [v for v in candidates if v <= 5_000_000]
        if plausible:
            return min(plausible)

        # Some files appear to serialize these fields as fixed-width integers.
        if len(data) in (2, 4):
            le = int.from_bytes(data, 'little', signed=False)
            if le <= 5_000_000:
                return le

        return min(candidates)

    # ======================================================================
    def _parse_var_struct_payload(self, payload: bytes):
        """Parse a variable-structure payload into (field_number, raw_value_bytes)."""
        stream = BytesIO(payload)
        fields = []

        try:
            field_cnt = self._read_varuint(stream)
        except EOFError:
            return fields

        for _ in range(field_cnt):
            key = self._read_varuint(stream)
            field_num = key >> 3
            val_len = key & 0x07

            if val_len == 7:
                val_len = self._read_varuint(stream)

            val = stream.read(val_len)
            fields.append((field_num, val))

        return fields

    # ======================================================================
    def _parseChannelInformation(self, chan_info_offset: int):
        """Parse header field 6 channel_information_array for each channel."""
        out = []

        with open(self.sonFile, 'rb') as file:
            file.seek(chan_info_offset)

            key = self._read_varuint(file)
            if key != 55:  # field 6 with length marker 7 -> 6<<3 + 7
                return out

            payload_len = self._read_varuint(file)
            payload = file.read(payload_len)

        stream = BytesIO(payload)

        try:
            chan_cnt = self._read_varuint(stream)
        except EOFError:
            return out

        for _ in range(chan_cnt):
            ch_payload = b''
            ch = {
                'channel_id': np.nan,
                'first_chunk_offset': np.nan,
                'transducer_port': np.nan,
                'start_freq_hz': np.nan,
                'end_freq_hz': np.nan,
                'channel_capabilities': np.nan,
            }

            # Each channel-info element is a variable structure.
            # Parse directly from stream by reading element field count and fields.
            try:
                elem_field_cnt = self._read_varuint(stream)
            except EOFError:
                break

            for _ in range(elem_field_cnt):
                fkey = self._read_varuint(stream)
                fnum = fkey >> 3
                flen = fkey & 0x07
                if flen == 7:
                    flen = self._read_varuint(stream)
                fval = stream.read(flen)

                if fnum == 0:
                    # data_info: varray of DataInfo structure(s)
                    info_stream = BytesIO(fval)
                    n = self._read_varuint(info_stream)
                    if n > 0:
                        for _ in range(n):
                            item_len = self._read_varuint(info_stream)
                            item = info_stream.read(item_len)
                            if item_len > 0:
                                ch['channel_id'] = self._decode_varuint_bytes(item)
                elif fnum == 1:
                    # first_chunk_offset: 8-byte little-endian ulong
                    if len(fval) == 8:
                        ch['first_chunk_offset'] = int.from_bytes(fval, 'little', signed=False)
                elif fnum == 2:
                    # prop_chan_info: varray of DpsChannelInformation struct(s)
                    dps_stream = BytesIO(fval)
                    dps_n = self._read_varuint(dps_stream)

                    for _ in range(dps_n):
                        dps_len = self._read_varuint(dps_stream)
                        dps_payload = dps_stream.read(dps_len)
                        dps_fields = self._parse_var_struct_payload(dps_payload)

                        for dnum, dval in dps_fields:
                            if dnum == 0:
                                ch['transducer_port'] = self._decode_varuint_bytes(dval)
                            elif dnum == 1:
                                tf_fields = self._parse_var_struct_payload(dval)
                                for tnum, tval in tf_fields:
                                    if tnum == 1:
                                        ch['start_freq_hz'] = self._decode_varuint_bytes(tval)
                                    elif tnum == 2:
                                        ch['end_freq_hz'] = self._decode_varuint_bytes(tval)
                            elif dnum == 2:
                                ch['channel_capabilities'] = self._decode_varuint_bytes(dval)

            out.append(ch)

        return out

    # ======================================================================
    def _parseBeamInfoPayload(self, payload: bytes):
        """Parse optional beam_info structure payload and return decoded values."""
        out = {}
        fields = self._parse_var_struct_payload(payload)

        for field_num, value in fields:
            if field_num == 0 and len(value) > 0:
                out['port_star_beam_angle'] = int(value[0])
            elif field_num == 1 and len(value) > 0:
                out['fore_aft_beam_angle'] = int(value[0])
            elif field_num == 2 and len(value) > 0:
                out['port_star_elem_angle'] = int(value[0])
            elif field_num == 3 and len(value) > 0:
                out['fore_aft_elem_angle'] = int(value[0])
            elif field_num == 5:
                # StructUnknown2 field 0 is strongly associated with port/star sign.
                su2_fields = self._parse_var_struct_payload(value)
                for su2_num, su2_val in su2_fields:
                    if su2_num == 0 and len(su2_val) == 4:
                        out['port_star_id'] = float(np.frombuffer(su2_val, dtype='<f4')[0])

        return out
    
    # ======================================================================
    def _getFileHeaderStruct(self):
        '''

        ffh: field - file header
        ffi: field - file information
        fci: field - channel information
        fcnt: field count
        '''


        # headBytes = 20480
        # firstHeadBytes = 35
        headStruct = [] 

        toCheck = {
            6:[('header_fcnt', '<u1')], #06: number of fields in header structure
            4:[('ffh_0', '<u1'), ('magic_number', '<u4')], #04: field 0 "magic_number", length 4
            10:[('ffh_1', '<u1'), ('format_version', '<u2')], #0a: field 1 "format_version", length 2
            20:[('ffh_2', '<u1'), ('channel_count', '<u4')], #14: field 2 "channel_count", length 4
            25:[('ffh_3', '<u1'), ('max_channel_count', '<u1')], #19: field 3 "max_channel_count", length 1
            47:[('ffh_4', '<u1'), ('ffh_4_actlen', '<u1'), ('ffi_fcnt', '<u1')], #2f: field 4 "file information", length 7; #actual length; #number of "file information" field 
            55: [('ffh_5', '<u1'), ('ffh_5_actlen', '<u2'), ('chan_cnt', '<u1'), ('fci_acnt', '<u1')], #37: "channel information", length 7; 
        }

        fileInfoToCheck = {
            2: [('ffi_0', '<u1'), ('unit_software_version', '<u2')], #02: "file information" field 0
            12: [('ffi_1', '<u1'), ('unit_id_type', '<u4')], #0c: "file information" field 1
            18: [('ffi_2', '<u1'), ('unit_product_number', '<u2')], #12: "file information" field 2    
            28: [('ffi_3', '<u1'), ('date_time_of_recording', '<u4')], #1c: "file information" field 3
        }

        # chanInfoToCheck = {
        #     3: [('fci_0_data_info', '<u1'), ('fci_0_acnt', '<u1'), ('fci_0_actlen', '<u1'), ('fci_channel_id', '<u1')], #03: Channel data info
        #     15: [('fci_1', '<u1'), ('fci_1_actlen', '<u1'), ('fci_first_chunk_offset', '<u8')], #0f: first chunk offset
        #     23: [('fci_2', '<u1'), ('fci_2_actlen', '<u2'), ('fci_2_acnt', '<u1'), ('fci_2_actlen2', '<u2')], #17 prop_chan_info

        
        # }

        # chanDataInfoToCheck = {

        # }

        file = open(self.sonFile, 'rb') # Open the file
        lastPos = 0 # Track last position in file

        foundChanInfo = False

        # while lastPos < firstHeadBytes - 1:
        while not foundChanInfo:
            # lastPos = file.tell()
            # print('lastPos:', lastPos)
            byte = self._fread_dat(file, 1, 'B')[0] # Decode the spacer byte

            if byte == 47:
                # File Information
                structDict = fileInfoToCheck

                for v in toCheck[byte]:
                    headStruct.append(v)

                length = self._fread_dat(file, 1, 'B')[0] # Decode the spacer byte
                field_cnt = self._fread_dat(file, 1, 'B')[0] # Decode the spacer byte

                fidx = 0
                while fidx < field_cnt:
                    byte = self._fread_dat(file, 1, 'B')[0] # Decode the spacer byte
                    if byte in structDict:
                        elen = 0
                        for v in structDict[byte]:

                            headStruct.append(v)
                            
                            # Get length of element
                            elen += (np.dtype(v[-1]).itemsize)

                        # Move forward elen amount
                        cpos = file.tell()
                        npos = cpos + elen - 1

                        file.seek(npos)
                        fidx += 1
                    else:
                        print('{} not in sonar header. Terminating.'.format(byte))
                        print('Offset: {}'.format(file.tell()))
                        sys.exit()

                # lastPos = headBytes
            
            # elif byte == 55:
            #     # File Information

            #     for v in toCheck[byte]:
            #         headStruct.append(v)

            #     length = self._fread_dat(file, 2, '<u2')[0] # Decode the spacer byte
            #     chan_cnt = self._fread_dat(file, 1, 'B')[0] # Decode the spacer byte

            #     chanidx = 0

            #     # Iterate each channel
            #     while chanidx < chan_cnt:

            #         field_cnt_0 = self._fread_dat(file, 1, 'B')[0] # Decode the spacer byte
            #         fidx_0 = 0

            #         # Iterate each field
            #         while fidx_0 < field_cnt_0:
            #             byte = self._fread_dat(file, 1, 'B')[0] # Decode the spacer byte

            #             if byte in chanInfoToCheck:
            #                 for v in chanInfoToCheck[byte]:
            #                     # Add chanidx to field name
            #                     field_name = '{}_{}'.format(v[0], chanidx)
            #                     v = (field_name, v[1])

            #                     headStruct.append(v)

            #             fidx_0 += 1

            #         chanidx += 1
                
            #     lastPos = headBytes

            elif byte == 55:
                foundChanInfo = True                            

            else:
                if byte in toCheck:
                    elen = 0
                    for v in toCheck[byte]:
                        headStruct.append(v)

                        # Get length of element
                        elen += (np.dtype(v[-1]).itemsize)
                    
                    # Move forward elen amount
                    cpos = file.tell()
                    npos = cpos + elen - 1

                    file.seek(npos)

            lastPos = file.tell()
            # print('lastPos:', lastPos)        

        return headStruct, lastPos-1
    
    
    ### Ping Header ###
    # ======================================================================
    def _parsePingHeader(self,):
        '''
        '''

        # Get the header struct
        self.son_struct, self.son_header_struct, self.record_body_header_len = self._getPingHeaderStruct()

        # Get the file length
        file_len = self.file_len

        # Initialize offset after file header
        i = self.headBytes

        # Open the file
        file = open(self.sonFile, 'rb')

        # Store contents in list
        header_dat_all = []

        # Decode ping header
        while i < file_len:

            # Get header data at offset i
            header_dat, cpos = self._getPingHeader(file, i)

            if header_dat:
                header_dat_all.append(header_dat)

            i = cpos

        # Convert to dataframe
        df = pd.DataFrame.from_dict(header_dat_all)

        # Convert fields
        df = self._doUnitConversion(df)

        # Do column name conversions to PINGMapper units
        df.rename(columns=self.garCols2PM, inplace=True)

        # Calculate speed & track distance (based on coords and time)
        df = self._calcSpeedTrkDist(df)

        # Drop negative son_offset
        df = df[df['son_offset'] > 0]


        # Test file to see outputs
        out_test = os.path.join(self.metaDir, 'All-Garmin-Sonar-MetaData.csv')
        df.to_csv(out_test, index=False)

        # Store in class
        self.header_dat = df

        return
    
    # ======================================================================
    def _getPingHeaderStruct(self, ):
        '''
        fpf: field - ping field
        fps: field - ping state
        '''

        headBytes = self.headBytes # Header length

        headStruct = [] 

        # pingHeaderToCheck = {
        #     6:[('header_fcnt', '<u1')], #06: number of fields in header structure
        #     4:[('fpf_0', '<u1'), ('magic_number', '<u4')], #04: field 0 "magic_number", length 4
        #     15:[('fpf_1', '<u1'), ('fpf_1_len', '<u1'), ('fpf_1_fcnt', '<u1'),
        #         ('fps_0', '<u1'), ('state', '<u1'),
        #         ('fps_1', '<u1'), ('data_info', '<u1'), ('data_info_cnt', '<u1'), ('data_info_len', '<u1'), ('channel_id', '<u1')], #0f: state data structure
        #     20:[('SP14', '<u1'), ('sequence_cnt', '<u4')], #14: sequence_count
        #     28:[('SP1c', '<u1'), ('data_crc', '<u4')], #1c: data crc
        #     34:[('SP22', '<u1'), ('data_size', '<u4')], #22: data size
        #     44:[('SP2c', '<u1'), ('recording_time_ms', '<u4')], #2c recording time offset
        # }

        # Record header (len==37)
        self.pingHeaderLen = pingHeaderLen = 37
        self.pingHeaderLenFirst = pingHeaderLenFirst = 49
        pingHeader = [
            ('header_fcnt', '<u1'),
            ('fpf_0', '<u1'), 
            ('magic_number', '<u4'),
            ('fpf_1', '<u1'), 
            ('fpf_1_len', '<u1'), 
            ('fpf_1_fcnt', '<u1'),
            ('fps_0', '<u1'), 
            ('state', '<u1'),
            ('fps_1', '<u1'), 
            # ('data_info', '<u1'), 
            ('data_info_cnt', '<u1'), 
            ('data_info_len', '<u1'), 
            ('channel_id', '<u1'),
            ('SP14', '<u1'), 
            ('sequence_cnt', '<u4'),
            ('SP1c', '<u1'), 
            ('data_crc', '<u4'),
            ('SP22', '<u1'),
            ('data_size', '<u2'),
            ('SP2c', '<u1'), 
            ('recording_time_ms', '<u4'),
            ('record_crc', '<u4')
        ]

        # pingBodyHeaderToCheck = {
        #     -1:('record_body_fcnt', '<u1'),
        #     1:[('SP1', '<u1'), ('channel_id_1', '<u1')], #01 channel_id
        #     11:[('SP0b', '<u1'), ('bottom_depth_unknown', '<u1'), ('bottom_depth', '<u2')], #0b bottom depth
        #     13:[('SP0d', '<u1'), ('unknown_sp0d', '<u4'), ('unknown_sp0d_1', '<u1')],
        #     18:[('SP12', '<u1'), ('unknown_sp12', '<u2')],
        #     19:[('SP13', '<u1'), ('drawn_bottom_depth_unknown', '<u1'), ('drawn_bottom_depth', '<u2')], #13 drawn bottom depth
        #     21:[('SP15', '<u1'), ('unknown_sp15', '<u4'), ('unknown_sp15_1', '<u1')],
        #     25:[('SP19', '<u1'), ('first_sample_depth', '<u1')], #19 first sample depth
        #     35:[('SP23', '<u1'), ('last_sample_depth_unknown', '<u1'), ('last_sample_depth', '<u2')], #23 last sample depth
        #     41:[('SP29', '<u1'), ('gain', '<u1')], #29 gain
        #     49:[('SP31', '<u1'), ('sample_status', '<u1')], #31 sample status
        #     60:[('SP3c', '<u1'), ('sample_cnt', '<u4')], #3c sample count
        #     65:[('SP41', '<u1'), ('shade_avail', '<u1')], #41 shade available
        #     76:[('SP4c', '<u1'), ('scposn_lat', '<u4')], #4c latitude
        #     84:[('SP54', '<u1'), ('scposn_lon', '<u4')], #54 longitude
        #     92:[('SP5c', '<u1'), ('water_temp', '<f4')], #5c temperature
        #     97:[('SP61', '<u1'), ('beam', '<u1')], #61 beam
        # }

        # magic number 86 DA E9 B7 == 3085556358
        # Ping headers always start at 20480

        # First and last state is 1. First time is the pingHeaderToCheck, forllowed
        ## by 16 CRC values, followed by first ping.

        # Beam 2 & 3 have extra unknown beam info (length 63)
        # Beam 1 & 4 sonar data starts immediately after 'beam'

        start_pos = lastPos = headBytes

        # Open file and move to offset
        file = open(self.sonFile, 'rb') # Open the file
        file.seek(start_pos)

        foundChanInfo = False

        headStruct = []

        # Start reading
        while not foundChanInfo:

            # Get the ping header (should be fixed structure)
            buffer = file.read(pingHeaderLen)

            # Read the data
            header = np.frombuffer(buffer, dtype=np.dtype(pingHeader))

            # Parse the data
            out_dict = {}
            for name, typ in header.dtype.fields.items():
                out_dict[name] = header[name][0].item()

            # Check if there is a record body
            if out_dict['state'] == 1: # no record body
                lastPos += pingHeaderLenFirst
                file.seek(lastPos)

            else:

                # # # Add pingheader
                # # for i in pingHeader:
                # #     headStruct.append(i)

                # # Get field count
                # field_cnt = self._fread_dat(file, 1, 'B')[0]
                # headStruct.append(pingBodyHeaderToCheck[-1])

                # if field_cnt > 13: # Only 13 known fields. Some beams have up to 15
                #     field_cnt = 13

                # fidx = 0
                # record_body_header_len = 1

                # while fidx < field_cnt:

                #     byte = self._fread_dat(file, 1, 'B')[0]

                #     if byte in pingBodyHeaderToCheck:
                #         elen = 0
                #         for v in pingBodyHeaderToCheck[byte]:
                #             headStruct.append(v)

                #             # Get length of element
                #             elen += (np.dtype(v[-1]).itemsize)

                #         # Move forward elen amount
                #         cpos = file.tell()
                #         npos = cpos + elen - 1

                #         record_body_header_len += elen

                #         file.seek(npos)

                #         fidx += 1

                #     else:
                #         print('{} not in sonar body. Terminating.'.format(byte))
                #         print('Offset: {}'.format(file.tell()))
                #         sys.exit()

                    foundChanInfo = True

        # self.son_header_struct = pingHeader
        # self.son_struct = headStruct

        # return headStruct, pingHeader, record_body_header_len
        return headStruct, pingHeader, 0

    # ======================================================================
    def _getPingHeader(self, file, i: int):

        # print('\n\n\n', i)

        # Get necessary attributes
        son_header_struct = self.son_header_struct
        pingHeaderLen = self.pingHeaderLen

        # head_struct = self.son_struct
        # record_body_header_len = self.record_body_header_len

        # Move to offset
        file.seek(i)

        # Get the ping header
        buffer = file.read(pingHeaderLen)

        # Read the data
        header = np.frombuffer(buffer, dtype=np.dtype(son_header_struct))

        out_dict = {}
        for name, typ in header.dtype.fields.items():
            out_dict[name] = header[name][0].item()

        # Check if there is a record body
        if out_dict['state'] != 2: # no record bod
            return False, i + self.pingHeaderLenFirst
        
        # # Get record body
        # # Get the ping header
        # buffer = file.read(record_body_header_len)

        # # Read the data
        # header = np.frombuffer(buffer, dtype=np.dtype(head_struct))

        # for name, typ in header.dtype.fields.items():
        #     out_dict[name] = header[name][0].item()

        # Variable structure so above doesn't work
        # Must determine structure ping by ping

        # Decode variable record body by field key/length instead of fixed count assumptions.
        rb_field_cnt = self._read_varuint(file)
        out_dict['record_body_fcnt'] = rb_field_cnt

        for _ in range(rb_field_cnt):
            key = self._read_varuint(file)
            field_num = key >> 3
            value_len = key & 0x07
            if value_len == 7:
                value_len = self._read_varuint(file)

            raw = file.read(value_len)

            if field_num == 0:
                out_dict['channel_id_1'] = self._decode_varuint_bytes(raw)
            elif field_num == 1:
                out_dict['bottom_depth'] = self._decode_depth_varint(raw)
            elif field_num == 2:
                out_dict['drawn_bottom_depth'] = self._decode_depth_varint(raw)
            elif field_num == 3:
                out_dict['first_sample_depth'] = self._decode_depth_varint(raw)
            elif field_num == 4:
                out_dict['last_sample_depth'] = self._decode_depth_varint(raw)
            elif field_num == 5 and len(raw) > 0:
                out_dict['gain'] = int(raw[0])
            elif field_num == 6:
                out_dict['sample_status'] = self._decode_varuint_bytes(raw)
            elif field_num == 7:
                out_dict['sample_cnt'] = int.from_bytes(raw, 'little', signed=False)
            elif field_num == 8 and len(raw) > 0:
                out_dict['shade_avail'] = int(raw[0])
            elif field_num == 9:
                out_dict['scposn_lat'] = int.from_bytes(raw, 'little', signed=True)
            elif field_num == 10:
                out_dict['scposn_lon'] = int.from_bytes(raw, 'little', signed=True)
            elif field_num == 11 and len(raw) == 4:
                out_dict['water_temp'] = float(np.frombuffer(raw, dtype='<f4')[0])
            elif field_num == 12:
                out_dict['beam'] = self._decode_varuint_bytes(raw)
            elif field_num == 13:
                out_dict.update(self._parseBeamInfoPayload(raw))
            elif field_num == 14:
                out_dict['interrogation_id'] = self._decode_varuint_bytes(raw)

            


        # Next ping header is from current position + ping_cnt
        # next_ping = file.tell() + out_dict['packet_size']
        next_ping = i + pingHeaderLen + out_dict['data_size'] + 12 #12 for magic number & crc

        out_dict['index'] = i

        sample_cnt = out_dict.get('sample_cnt', 0)
        out_dict['son_offset'] = (out_dict['data_size']) - (sample_cnt*2) + self.pingHeaderLen

        # out_dict['son_offset'] = record_body_header_len+1
 
        return out_dict, next_ping
    
    
    ### Ping Header Conversions ###
    # ======================================================================
    def _calcSpeedTrkDist(self, df: pd.DataFrame, jump_thresh: float=1.0):

        x = df['e'].to_numpy()
        y = df['n'].to_numpy()
        t = df['time_s'].to_numpy()# / 1000
        t = np.diff(t)
        ds = np.zeros((len(x)))

        x1 = x[:-1]
        y1 = y[:-1]
        x2 = x[1:]
        y2 = y[1:]

        d = np.sqrt((x2 - x1)**2 + (y2 - y1)**2)
        ds[1:] = d
        df['dist'] = ds

        # Calculate speed
        s = np.full_like(t, np.nan, dtype='float64')
        np.divide(ds[1:], t, out=s, where=t != 0)
        s = np.append(s[0], s)
        s = pd.Series(s).ffill().bfill().to_numpy()


        # Assume constant speed for nan's. Need to interpolate.
        s = pd.Series(s, index=df.index)
        s.replace(0, np.nan, inplace=True)  # Replace 0 with NaN for interpolation
        s.interpolate(method='linear', inplace=True)

        # Accumulate distance
        ds = np.cumsum(ds)

        # Store
        df['speed_ms'] = np.around(s, 1)
        df['trk_dist'] = ds
        # df['speed_ms'] = df['speed_ms'].fillna(0)



        return df
    
    # ======================================================================
    def _doUnitConversion(self, df: pd.DataFrame):

        #####################
        # Convert depth units
        def decode_varint(data):
            """Decodes a varint from a byte string.

            Args:
                data: A byte string containing the varint.

            Returns:
                The decoded integer value.
            """
            result = 0
            shift = 0
            for byte in data:
                result |= (byte & 0x7f) << shift
                shift += 7
                if not (byte & 0x80):  # If MSB is 0, end of varint
                    break
            return result

        # Garmin depth-like fields are stored in millimeters. Convert to meters.
        cols_to_convert = ['bottom_depth', 'drawn_bottom_depth', 'last_sample_depth']
        for col in cols_to_convert:

            if col in df.columns:
                df[col] = df[col].apply(lambda x: decode_varint(x) if isinstance(x, bytes) else x)
                # Garmin uses sentinel-style out-of-range values in some files.
                df[col] = df[col].where((df[col] >= 0) & (df[col] < 0xFFFFFF00), np.nan)
                df[col] = df[col].astype(float) / 1000.0

        if 'first_sample_depth' in df.columns:
            df['first_sample_depth'] = df['first_sample_depth'].where(
                (df['first_sample_depth'] >= 0) & (df['first_sample_depth'] < 0xFFFFFF00),
                np.nan,
            )
            df['first_sample_depth'] = df['first_sample_depth'].astype(float) / 1000.0



        ##############
        # Convert time
        # Garmin uses 0000 December 31, 1989 as start time

        start_date = datetime(1989, 12, 31, 0, 0, 0, tzinfo=timezone.utc)

        custom_unix_time = start_date + timedelta(seconds=self.file_header['date_time_of_recording'])

        custom_unix_time = custom_unix_time.timestamp()

        df['recording_time_ms'] /= 1000  # Convert ms to seconds

        custom_unix_time = custom_unix_time + (df['recording_time_ms'])

        df['caltime'] = pd.to_datetime(custom_unix_time, unit='s', utc=True)

        df['date'] = df['caltime'].dt.date
        df['time'] = df['caltime'].dt.time # Time in utc

        df.drop(columns=['caltime'], inplace=True)


        ##################################
        # Calculate latitude and longitude
        # df['lat'] = df['scposn_lat'] * 360 / (1<<32)
        # df['lon'] = df['scposn_lon'] * 360 / (1<<32)

        # df['lat'] = df['scposn_lat'].astype('float64') * (180.0 / (2**31))
        # df['lon'] = df['scposn_lon'].astype('float64') * (180.0 / (2**31))

        df['lat'] = df['scposn_lat'] * 360 / (1 << 32)
        df['lon'] = df['scposn_lon'] * 360 / (1 << 32)

        df['lat'] = df['lat'].apply(lambda x: x - 360 if x > 180 else x)
        df['lon'] = df['lon'].apply(lambda x: x - 360 if x > 180 else x)

        # Do filtering
        df = filterGPS(df)




        # print('\n\nConverted lat lon:')
        # print(df[['lat', 'lon']].describe())

        # import matplotlib.pyplot as plt

        # plt.scatter(df['scposn_lon'], df['scposn_lat'], s=1, alpha=0.5)
        # plt.title('Vessel Track')
        # plt.xlabel('Longitude')
        # plt.ylabel('Latitude')
        # plt.grid(True)
        # plt.show()

        # plt.figure(figsize=(10, 6))
        # plt.scatter(df['lon'], df['lat'], s=1, alpha=0.5, label='Cleaned')
        # plt.title('GPS Track After Percentile Filtering')
        # plt.xlabel('Longitude')
        # plt.ylabel('Latitude')
        # plt.grid(True)
        # plt.legend()
        # plt.show()



        # Determine epsg code
        self.humDat['epsg'] = "EPSG:"+str(int(float(self._convert_wgs_to_utm(df['lon'][0], df['lat'][0]))))
        self.humDat['wgs'] = "EPSG:4326"

        # Configure re-projection function
        self.trans = pyproj.Proj(self.humDat['epsg'])

        # Reproject lat/lon to UTM zone
        e, n = self.trans(df['lon'], df['lat'])
        df['e'] = e
        df['n'] = n


        #########################
        # Calculate COG (heading)
        ## Garmin does not appear to store heading....
        heading = self._getCOG(df)
        # self._getBearing() returns n-1 values because last ping can't
        ## have a COG value.  We will duplicate the last COG value and use it for
        ## the last ping.
        last = heading[-1]
        heading = np.append(heading, last)
        df['instr_heading'] = heading # Store COG in sDF

        # Replace 0 with NaN, interpolate, then fill any remaining NaN (e.g., with nearest valid value)
        df['instr_heading'] = np.around(df['instr_heading'].replace(0, np.nan).interpolate().bfill().ffill(), 1)

        # Add transect number (for aoi processing)
        df['transect'] = 0

        # Calculate pixel size [m]  *** ....MAYBE.... ***
        df['pixM'] = (df['last_sample_depth'] - df['first_sample_depth']) / df['sample_cnt']

        return df
    
    #===========================================
    def _getCOG(self,
                df,
                lon = 'lon',
                lat = 'lat'):
        '''
        Calculates course over ground (COG) from a set of coordinates.  Since the
        last coordinate pair cannot have a COG value, the length of the returned
        array is len(n-1) where n == len(df).

        ----------
        Parameters
        ----------
        df : DataFrame
            DESCRIPTION - Pandas dataframe with geographic coordinates of sonar
                          records.
        lon : str : [Default='lons']
            DESCRIPTION - DataFrame column name for longitude coordinates.
        lat : str : [Default='lats']
            DESCRIPTION - DataFrame column name for latitude coordinates.

        ----------------------------
        Required Pre-processing step
        ----------------------------
        Called from self._interpTrack()

        -------
        Returns
        -------
        Numpy array of COG values.

        --------------------
        Next Processing Step
        --------------------
        Return to self._interpTrack()
        '''
        # COG calculation will be calculated on numpy arrays for speed.  Since
        ## COG is calculated from one point to another (pntA -> pntB), we need
        ## to store pntA values, beginning with the first value and ending at
        ## second to last value, in one array and pntB values, beginning at second
        ## value and ending at last value, in another array.  We can then use
        ## vector algebra to efficiently calculate COG.

        # Prepare pntA values [0:n-1]
        lonA = df[lon].to_numpy() # Store longitude coordinates in numpy array
        latA = df[lat].to_numpy() # Store longitude coordinates in numpy array
        lonA = lonA[:-1] # Omit last coordinate
        latA = latA[:-1] # Omit last coordinate
        pntA = [lonA,latA] # Store in array of arrays

        # Prepare pntB values [0+1:n]
        lonB = df[lon].to_numpy() # Store longitude coordinates in numpy array
        latB = df[lat].to_numpy() # Store longitude coordinates in numpy array
        lonB = lonB[1:] # Omit first coordinate
        latB = latB[1:] # Omit first coordinate
        pntB = [lonB,latB] # Store in array of arrays

        # Convert latitude values into radians
        lat1 = np.deg2rad(pntA[1])
        lat2 = np.deg2rad(pntB[1])

        diffLong = np.deg2rad(pntB[0] - pntA[0]) # Calculate difference in longitude then convert to degrees
        bearing = np.arctan2(np.sin(diffLong) * np.cos(lat2), np.cos(lat1) * np.sin(lat2) - (np.sin(lat1) * np.cos(lat2) * np.cos(diffLong))) # Calculate bearing in radians

        db = np.degrees(bearing) # Convert radians to degrees
        db = (db + 360) % 360 # Ensure degrees in range 0-360

        return db
    
    # ======================================================================
    def _convert_wgs_to_utm(self, lon: float, lat: float):
        """
        This function estimates UTM zone from geographic coordinates
        see https://stackoverflow.com/questions/40132542/get-a-cartesian-projection-accurate-around-a-lat-lng-pair
        """
        utm_band = str((np.floor((lon + 180) / 6 ) % 60) + 1)
        if len(utm_band) == 1:
            utm_band = '0'+utm_band
        if lat >= 0:
            epsg_code = '326' + utm_band
        else:
            epsg_code = '327' + utm_band
        return epsg_code
    
    
    ### Format to PINGMapper ###
    # ======================================================================
    def _recalcRecordNum(self):

        df = self.header_dat

        # Reset index and recalculate record num
        ## Record num is unique for each ping across all sonar beams
        df = df.reset_index(drop=True)
        df['record_num'] = df.index

        self.header_dat = df
        return
    

    # ======================================================================
    def _splitBeamsToCSV(self):

        beam_set = {}

        # Dictionary to store necessary attributes for PING-Mapper
        self.beamMeta = beamMeta = {}

        # Get df
        df = self.header_dat

        # Prefer explicit in-record beam values and use port_star_id sign to disambiguate sidescan.
        beam_name_by_id = {
            1: 'ds_hifreq',
            2: 'ss_port',
            3: 'ss_star',
            4: 'ds_vhifreq',
        }

        chan_ids = sorted(df['channel_id'].dropna().unique())

        for chan_id in chan_ids:
            g = df[df['channel_id'] == chan_id]

            beam_id = np.nan
            if 'beam' in g.columns:
                b = g['beam'].dropna()
                b = b[b.isin([1, 2, 3, 4])]
                if len(b) > 0:
                    beam_id = int(b.mode().iloc[0])

            if np.isnan(beam_id) and isinstance(getattr(self, 'channel_info', None), list):
                # Conservative fallback from channel header metadata when beam is not present.
                match = [c for c in self.channel_info if c.get('channel_id', np.nan) == chan_id]
                if len(match) == 1:
                    cinfo = match[0]
                    sf = cinfo.get('start_freq_hz', np.nan)
                    ef = cinfo.get('end_freq_hz', np.nan)
                    if np.isfinite(sf) and np.isfinite(ef):
                        fkhz = (sf + ef) / 2000.0
                        if fkhz >= 700:
                            beam_id = 4
                        elif fkhz >= 300:
                            beam_id = 2
                        else:
                            beam_id = 1

            if np.isnan(beam_id):
                beam_set[chan_id] = ('unknown', -1)
            else:
                beam_set[chan_id] = (beam_name_by_id.get(int(beam_id), 'unknown'), int(beam_id))

        # Use sign of port_star_id to force sidescan orientation when available.
        if 'port_star_id' in df.columns:
            for chan_id in chan_ids:
                g = df[df['channel_id'] == chan_id]
                ps = g['port_star_id'].dropna()
                if len(ps) == 0:
                    continue

                ps_med = float(np.median(ps))
                if ps_med < 0:
                    beam_set[chan_id] = ('ss_port', 2)
                elif ps_med > 0:
                    beam_set[chan_id] = ('ss_star', 3)

        # Final fallback if still unresolved.
        unresolved = [k for k, v in beam_set.items() if v[1] == -1]
        if len(unresolved) > 0:
            ordered = sorted(chan_ids)
            for idx, chan_id in enumerate(ordered):
                if chan_id not in unresolved:
                    continue
                if idx == 0:
                    beam_set[chan_id] = ('ds_hifreq', 1)
                elif idx == 1:
                    beam_set[chan_id] = ('ss_port', 2)
                elif idx == 2:
                    beam_set[chan_id] = ('ss_star', 3)
                else:
                    beam_set[chan_id] = ('ds_vhifreq', 4)


        # Iterate each beam
        for beam, group in df.groupby('channel_id'):
            meta = {}
            
            # Get Garmin beam to Humminbird beam
            humBeamName, humBeamint = beam_set[beam]
            humBeam = 'B00'+str(humBeamint)
            meta['beamName'] = humBeamName
            meta['beam'] = humBeam
            group = group.copy()
            group['beam'] = humBeamint

            # # Set pixM based on side scan
            # if humBeamint == 2 or humBeamint == 3:
            #     self.pixM = group['pixM'].iloc[0]

            # Store sonFile
            meta['sonFile'] = self.sonFile

            # Drop columns
            cols2Drop = ['magic_number']
            cols = group.columns
            cols2Drop += [c for c in cols if 'fp' in c]
            cols2Drop += [c for c in cols if 'SP' in c]
            cols2Drop += [c for c in cols if 'su' in c]
            group.drop(columns=cols2Drop, inplace=True)

            # Add chunk_id
            group = self._getChunkID(group)

            # Save csv
            outCSV = '{}_{}_meta.csv'.format(humBeam, meta['beamName'])
            outCSV = os.path.join(self.metaDir, outCSV)
            group.to_csv(outCSV, index=False)

            meta['metaCSV'] = outCSV

            # Store the beams metadata
            beamMeta[humBeam] = meta

        return
    

    # ======================================================================
    def _getChunkID(self, df: pd.DataFrame):

        df.reset_index(drop=True, inplace=True)

        df['chunk_id'] = int(-1)

        chunk = 0
        start_idx = chunk
        end_idx = self.nchunk

        while start_idx < len(df):

            df.iloc[start_idx:end_idx, df.columns.get_loc('chunk_id')] = int(chunk)

            chunk += 1
            start_idx = end_idx
            end_idx += self.nchunk

        # Update last chunk if too small (for rectification)
        lastChunk = df[df['chunk_id'] == chunk]
        if len(lastChunk) <= self.nchunk/2:
            df.loc[df['chunk_id'] == chunk, 'chunk_id'] = chunk-1


        return df





    # ======================================================================
    def __str__(self):
        '''
        Generic print function to print contents of sonObj.
        '''
        output = "Lowrance Class Contents"
        output += '\n\t'
        output += self.__repr__()
        temp = vars(self)
        for item in temp:
            output += '\n\t'
            output += "{} : {}".format(item, temp[item])
        return output


import os
import struct
from datetime import datetime

import numpy as np
import pandas as pd
import pyproj


class xtf(object):

    def __init__(self, inFile: str, nchunk: int = 0, exportUnknown: bool = False):
        self.humFile = None
        self.isOnix = 0
        self.sonFile = inFile
        self.nchunk = nchunk
        self.exportUnknown = exportUnknown

        self.file_header_size = 1024
        self.ping_header_size = 256
        self.ping_chan_header_size = 64

        self.humDat = {}
        self.headBytes = 0
        self.son8bit = False
        self.sample_dtype = '<u2'
        self.nav_units = None

    def _getFileLen(self):
        self.file_len = os.path.getsize(self.sonFile)
        return

    def _parseFileHeader(self):
        with open(self.sonFile, 'rb') as file:
            base = file.read(1024)

        if len(base) < 1024:
            raise ValueError('Invalid XTF file: file header shorter than 1024 bytes.')

        nav_units = struct.unpack_from('<H', base, 164)[0]
        self.nav_units = int(nav_units)
        sonar_channels = struct.unpack_from('<H', base, 166)[0]

        if sonar_channels > 6:
            extra = sonar_channels - 6
            extra_blocks = int(np.ceil(extra / 8.0))
            self.file_header_size = 1024 + (extra_blocks * 1024)
        else:
            self.file_header_size = 1024

        with open(self.sonFile, 'rb') as file:
            header = file.read(self.file_header_size)

        chaninfo = self._parse_chaninfo(header, sonar_channels)

        self.file_header = {
            'nav_units': nav_units,
            'number_of_sonar_channels': sonar_channels,
            'file_header_size': self.file_header_size,
        }
        self.chaninfo = chaninfo

        out_file = os.path.join(self.metaDir, 'DAT_meta.csv')
        pd.DataFrame.from_dict(self.file_header, orient='index').T.to_csv(out_file, index=False)
        self.datMetaFile = out_file

        return

    def _parse_chaninfo(self, header_bytes: bytes, sonar_channels: int):
        chaninfo = {}
        base_offset = 256
        chan_size = 128

        for idx in range(sonar_channels):
            start = base_offset + idx * chan_size
            end = start + chan_size
            if end > len(header_bytes):
                break

            block = header_bytes[start:end]
            bytes_per_sample = struct.unpack_from('<H', block, 6)[0]
            frequency = struct.unpack_from('<f', block, 32)[0]
            sample_format = block[74]
            type_of_channel = block[0]
            sub_channel_number = block[1]

            chaninfo[idx] = {
                'type_of_channel': type_of_channel,
                'sub_channel_number': sub_channel_number,
                'bytes_per_sample': bytes_per_sample,
                'frequency': frequency,
                'sample_format': sample_format,
            }

        return chaninfo

    def _parsePingHeader(self):
        with open(self.sonFile, 'rb') as file:
            file_len = self.file_len
            i = self.file_header_size
            rows = []

            while i + self.ping_header_size <= file_len:
                file.seek(i)
                header = file.read(self.ping_header_size)

                if len(header) < self.ping_header_size:
                    break

                magic = struct.unpack_from('<H', header, 0)[0]
                if magic != 0xFACE:
                    i += 1
                    continue

                header_type = header[2]
                num_chans_to_follow = struct.unpack_from('<H', header, 4)[0]
                num_bytes_this_record = struct.unpack_from('<I', header, 10)[0]

                if num_bytes_this_record <= 0:
                    i += 1
                    continue

                if i + num_bytes_this_record > file_len:
                    break

                if header_type == 0 and num_chans_to_follow > 0:
                    rows.extend(self._parse_sonar_record(i, header, num_chans_to_follow, num_bytes_this_record, file))

                i += num_bytes_this_record

        df = pd.DataFrame.from_dict(rows)

        if len(df) == 0:
            raise ValueError('No XTF sonar ping packets (HeaderType 0) were parsed.')

        df = self._split_combined_sidescan(df)
        df = self._sync_port_star_metadata(df)

        df.sort_values(by=['time_s', 'beam'], inplace=True)
        df.reset_index(drop=True, inplace=True)

        df = self._doUnitConversion(df)
        df = self._calcTrkDistTS(df)

        df['record_num'] = np.arange(len(df), dtype=np.int64)

        out_test = os.path.join(self.metaDir, 'All-XTF-Sonar-MetaData.csv')
        df.to_csv(out_test, index=False)

        self.header_dat = df
        return

    def _split_combined_sidescan(self, df: pd.DataFrame):
        return self._split_combined_sidescan_group(df)

    def _split_combined_sidescan_group(self, df: pd.DataFrame):
        beams = set(df['beam'].dropna().astype(int).unique().tolist())

        if 2 in beams and 3 in beams:
            return df

        if beams == {2} or beams == {3}:
            source_beam = 2 if 2 in beams else 3
            src = df[df['beam'] == source_beam].copy()

            if len(src) == 0:
                return df

            port = src.copy()
            star = src.copy()

            port['beam'] = 2
            star['beam'] = 3

            half_cnt = (src['ping_cnt'] / 2).astype(int)
            port['ping_cnt'] = half_cnt
            star['ping_cnt'] = half_cnt

            bytes_per_sample = src['bytes_per_sample'].fillna(1).astype(int)
            star['son_offset'] = src['son_offset'] + (half_cnt * bytes_per_sample)

            if 'channel_number' in port.columns:
                port['channel_number'] = 0
            if 'channel_number' in star.columns:
                star['channel_number'] = 1

            combined = pd.concat([port, star], ignore_index=True)
            combined.sort_values(by=['time_s', 'beam'], inplace=True)
            combined.reset_index(drop=True, inplace=True)
            return combined

        return df

    def _sync_port_star_metadata(self, df: pd.DataFrame):
        if 'beam' not in df.columns:
            return df

        if 'freq_band' not in df.columns:
            return self._sync_port_star_metadata_group(df)

        grouped = []
        for _, group in df.groupby('freq_band', dropna=False):
            grouped.append(self._sync_port_star_metadata_group(group))

        out = pd.concat(grouped, ignore_index=True)
        out.sort_values(by=['time_s', 'beam'], inplace=True)
        out.reset_index(drop=True, inplace=True)
        return out

    def _invalid_geom_mask(self, df: pd.DataFrame):
        pix = pd.to_numeric(df['pixM'], errors='coerce') if 'pixM' in df.columns else pd.Series(np.nan, index=df.index)
        spp = pd.to_numeric(df['seconds_per_ping'], errors='coerce') if 'seconds_per_ping' in df.columns else pd.Series(np.nan, index=df.index)
        ping = pd.to_numeric(df['ping_cnt'], errors='coerce') if 'ping_cnt' in df.columns else pd.Series(np.nan, index=df.index)

        return (
            pix.isna() |
            (pix <= 1e-4) |
            spp.isna() |
            (spp <= 1e-6) |
            ping.isna() |
            (ping <= 0) |
            (ping > 1_000_000)
        )

    def _repair_invalid_metadata_rows(self, df: pd.DataFrame):
        if len(df) == 0:
            return df

        out = df.copy()
        invalid = self._invalid_geom_mask(out)
        if not invalid.any():
            return out

        key = 'ping_number' if 'ping_number' in out.columns else 'time_s'
        out.sort_values(by=[key], inplace=True)
        invalid = self._invalid_geom_mask(out)

        repair_fields = ['pixM', 'seconds_per_ping', 'f', 'f_min', 'f_max', 'bytes_per_sample', 'ping_cnt']
        for field in repair_fields:
            if field not in out.columns:
                continue

            series = pd.to_numeric(out[field], errors='coerce')
            series = series.where(~invalid, np.nan)
            series = series.ffill().bfill()

            if field in ['bytes_per_sample', 'ping_cnt']:
                out[field] = np.round(series).astype(int)
            else:
                out[field] = series

        out.reset_index(drop=True, inplace=True)
        return out

    def _sync_port_star_metadata_group(self, df: pd.DataFrame):
        beams = set(df['beam'].dropna().astype(int).unique().tolist())
        if not ({2, 3}.issubset(beams)):
            return self._repair_invalid_metadata_rows(df)

        port = df[df['beam'] == 2].copy()
        star = df[df['beam'] == 3].copy()

        if len(port) == 0 or len(star) == 0:
            return df

        key = 'ping_number' if 'ping_number' in df.columns else 'index'

        port_lookup = port.sort_values(by=[key]).drop_duplicates(subset=[key], keep='first').set_index(key)
        star_lookup = star.sort_values(by=[key]).drop_duplicates(subset=[key], keep='first').set_index(key)

        star_needs_geom = self._invalid_geom_mask(star)
        if 'channel_number' in star.columns:
            star_chan = pd.to_numeric(star['channel_number'], errors='coerce')
            star_needs_geom |= (~star_chan.isin([0, 1, 2, 3, 4, 5]))

        src_ping_for_star = pd.to_numeric(star[key].map(port_lookup['ping_cnt']), errors='coerce')
        dst_ping_star = pd.to_numeric(star['ping_cnt'], errors='coerce')
        ping_mismatch_star = np.isfinite(src_ping_for_star) & np.isfinite(dst_ping_star) & (
            (dst_ping_star > (src_ping_for_star * 2.0)) |
            (dst_ping_star < (src_ping_for_star * 0.5))
        )

        src_spp_for_star = pd.to_numeric(star[key].map(port_lookup['seconds_per_ping']), errors='coerce')
        dst_spp_star = pd.to_numeric(star['seconds_per_ping'], errors='coerce')
        spp_mismatch_star = np.isfinite(src_spp_for_star) & np.isfinite(dst_spp_star) & (
            (dst_spp_star > max(5.0, np.nanmax(src_spp_for_star) * 5.0 if np.isfinite(np.nanmax(src_spp_for_star)) else 5.0)) |
            (dst_spp_star > (src_spp_for_star * 5.0)) |
            (dst_spp_star < (src_spp_for_star * 0.2))
        )

        star_needs_geom |= ping_mismatch_star | spp_mismatch_star

        for idx in star[star_needs_geom].index:
            k = star.loc[idx, key]
            if k not in port_lookup.index:
                continue
            src = port_lookup.loc[k]

            for field in ['pixM', 'seconds_per_ping', 'f', 'f_min', 'f_max', 'bytes_per_sample', 'ping_cnt']:
                if field in star.columns and field in src.index:
                    star.at[idx, field] = src[field]

            if 'channel_number' in star.columns:
                star.at[idx, 'channel_number'] = 1

        port_needs_geom = self._invalid_geom_mask(port)
        if 'channel_number' in port.columns:
            port_chan = pd.to_numeric(port['channel_number'], errors='coerce')
            port_needs_geom |= (~port_chan.isin([0, 1, 2, 3, 4, 5]))

        src_ping_for_port = pd.to_numeric(port[key].map(star_lookup['ping_cnt']), errors='coerce')
        dst_ping_port = pd.to_numeric(port['ping_cnt'], errors='coerce')
        ping_mismatch_port = np.isfinite(src_ping_for_port) & np.isfinite(dst_ping_port) & (
            (dst_ping_port > (src_ping_for_port * 2.0)) |
            (dst_ping_port < (src_ping_for_port * 0.5))
        )

        src_spp_for_port = pd.to_numeric(port[key].map(star_lookup['seconds_per_ping']), errors='coerce')
        dst_spp_port = pd.to_numeric(port['seconds_per_ping'], errors='coerce')
        spp_mismatch_port = np.isfinite(src_spp_for_port) & np.isfinite(dst_spp_port) & (
            (dst_spp_port > max(5.0, np.nanmax(src_spp_for_port) * 5.0 if np.isfinite(np.nanmax(src_spp_for_port)) else 5.0)) |
            (dst_spp_port > (src_spp_for_port * 5.0)) |
            (dst_spp_port < (src_spp_for_port * 0.2))
        )

        port_needs_geom |= ping_mismatch_port | spp_mismatch_port

        for idx in port[port_needs_geom].index:
            k = port.loc[idx, key]
            if k not in star_lookup.index:
                continue
            src = star_lookup.loc[k]
            for field in ['pixM', 'seconds_per_ping', 'f', 'f_min', 'f_max', 'bytes_per_sample', 'ping_cnt']:
                if field in port.columns and field in src.index:
                    port.at[idx, field] = src[field]

            if 'channel_number' in port.columns:
                port.at[idx, 'channel_number'] = 0

        port = self._repair_invalid_metadata_rows(port)
        star = self._repair_invalid_metadata_rows(star)

        out = pd.concat([port, star], ignore_index=True)
        out.sort_values(by=['time_s', 'beam'], inplace=True)
        out.reset_index(drop=True, inplace=True)
        return out

    def _normalize_channel_number(self, channel_number: int, chan_idx: int):
        try:
            channel_number = int(channel_number)
        except Exception:
            channel_number = -1

        if channel_number in (0, 1, 2, 3, 4, 5):
            return channel_number

        # Some Klein files store corrupted/non-canonical channel IDs in packet
        # channel headers while channel order (chan_idx) remains stable.
        if 0 <= int(chan_idx) <= 5:
            return int(chan_idx)

        return None

    def _parse_sonar_record(self, record_start: int, header: bytes, num_chans_to_follow: int, record_bytes: int, file):
        year = struct.unpack_from('<H', header, 14)[0]
        month = header[16]
        day = header[17]
        hour = header[18]
        minute = header[19]
        second = header[20]
        hsecond = header[21]

        ping_number = struct.unpack_from('<I', header, 28)[0]
        sensor_speed_kn = struct.unpack_from('<f', header, 152)[0]
        sensor_y = struct.unpack_from('<d', header, 160)[0]
        sensor_x = struct.unpack_from('<d', header, 168)[0]
        sensor_depth = struct.unpack_from('<f', header, 192)[0]
        sensor_altitude = struct.unpack_from('<f', header, 196)[0]
        sensor_pitch = struct.unpack_from('<f', header, 204)[0]
        sensor_roll = struct.unpack_from('<f', header, 208)[0]
        sensor_heading = struct.unpack_from('<f', header, 212)[0]
        sensor_yaw = struct.unpack_from('<f', header, 220)[0]

        rows = []

        sample_offset = self.ping_header_size + (num_chans_to_follow * self.ping_chan_header_size)

        try:
            dt = datetime(year, max(month, 1), max(day, 1), hour, minute, second)
            time_s = dt.timestamp() + (hsecond / 100.0)
        except Exception:
            time_s = float(ping_number)

        lat, lon, e, n = self._decode_position(sensor_x, sensor_y, self.nav_units)

        ref_ping_cnt = None

        for chan_idx in range(num_chans_to_follow):
            if int(sample_offset) >= int(record_bytes):
                break

            chan_off = self.ping_header_size + (chan_idx * self.ping_chan_header_size)
            file.seek(record_start + chan_off)
            chan_header = file.read(self.ping_chan_header_size)

            if len(chan_header) < self.ping_chan_header_size:
                break

            channel_number = struct.unpack_from('<H', chan_header, 0)[0]
            slant_range = struct.unpack_from('<f', chan_header, 4)[0]
            seconds_per_ping = struct.unpack_from('<f', chan_header, 20)[0]
            frequency = struct.unpack_from('<H', chan_header, 26)[0]
            num_samples = struct.unpack_from('<I', chan_header, 42)[0]
            try:
                num_samples = int(num_samples)
            except Exception:
                num_samples = 0

            chan_cfg = self.chaninfo.get(chan_idx, {})
            type_of_channel = chan_cfg.get('type_of_channel', None)
            bytes_per_sample = max(int(chan_cfg.get('bytes_per_sample', 1)), 1)

            remaining_payload_bytes = max(int(record_bytes) - int(sample_offset), 0)
            max_samples = remaining_payload_bytes // bytes_per_sample

            if type_of_channel is not None and type_of_channel not in [1, 2]:
                skip_samples = max(0, min(int(num_samples), int(max_samples)))
                sample_offset += int(skip_samples) * int(bytes_per_sample)
                continue

            channel_number = self._normalize_channel_number(channel_number, chan_idx)
            if channel_number is None:
                skip_samples = max(0, min(int(num_samples), int(max_samples)))
                sample_offset += int(skip_samples) * int(bytes_per_sample)
                continue

            chan_freq = float(chan_cfg.get('frequency', frequency)) if chan_cfg.get('frequency', frequency) is not None else np.nan

            beam = self._map_beam(channel_number, type_of_channel, chan_idx)
            freq_band = self._map_freq_band(channel_number, chan_idx)

            ping_cnt = int(num_samples)
            if ping_cnt < 0:
                ping_cnt = 0
            if max_samples <= 0:
                ping_cnt = 0
            elif ping_cnt > max_samples:
                ping_cnt = int(max_samples)

            if ref_ping_cnt is None and ping_cnt > 0:
                ref_ping_cnt = int(ping_cnt)

            if (
                ref_ping_cnt is not None and
                ref_ping_cnt > 0 and
                (num_samples <= 0 or int(num_samples) > int(max_samples) or ping_cnt > (ref_ping_cnt * 2))
            ):
                ping_cnt = int(min(ref_ping_cnt, max_samples))

            if ping_cnt <= 0:
                continue

            pix_m = np.nan
            if slant_range > 0:
                pix_m = slant_range / ping_cnt

            sample_format = int(chan_cfg.get('sample_format', 0))
            if bytes_per_sample == 1:
                self.son8bit = True
                self.sample_dtype = '>u1'
            elif bytes_per_sample == 2:
                self.son8bit = False
                self.sample_dtype = '<u2'
            elif bytes_per_sample == 4:
                self.son8bit = False
                # XTF sample format 5 is commonly float32 amplitudes (e.g., Klein).
                if sample_format == 5:
                    self.sample_dtype = '<f4'
                else:
                    self.sample_dtype = '<u4'
            else:
                self.son8bit = False
                self.sample_dtype = '<u2'

            sensor_depth_m = float(sensor_depth) if np.isfinite(sensor_depth) and float(sensor_depth) > 0 else np.nan
            sensor_altitude_m = float(sensor_altitude) if np.isfinite(sensor_altitude) and float(sensor_altitude) > 0 else np.nan

            dep_m = sensor_depth_m
            if not np.isfinite(dep_m) and np.isfinite(sensor_altitude_m):
                dep_m = sensor_altitude_m

            row = {
                'index': int(record_start),
                'son_offset': int(sample_offset),
                'ping_number': int(ping_number),
                'time_s': float(time_s),
                'beam': int(beam),
                'channel_number': int(channel_number),
                'ping_cnt': ping_cnt,
                'bytes_per_sample': int(bytes_per_sample),
                'f': float(chan_freq) if np.isfinite(chan_freq) else np.nan,
                'f_min': float(chan_freq) if np.isfinite(chan_freq) else np.nan,
                'f_max': float(chan_freq) if np.isfinite(chan_freq) else np.nan,
                'freq_band': freq_band,
                'pixM': float(pix_m) if np.isfinite(pix_m) else np.nan,
                'speed_ms': float(sensor_speed_kn) * 0.514444,
                'inst_dep_m': sensor_depth_m,
                'instr_heading': float(sensor_heading),
                'pitch': float(sensor_pitch) if np.isfinite(sensor_pitch) else np.nan,
                'roll': float(sensor_roll) if np.isfinite(sensor_roll) else np.nan,
                'yaw': float(sensor_yaw) if np.isfinite(sensor_yaw) else np.nan,
                'dep_m': dep_m,
                'altitude': sensor_altitude_m,
                'lat': lat,
                'lon': lon,
                'e': e,
                'n': n,
                'seconds_per_ping': float(seconds_per_ping) if np.isfinite(seconds_per_ping) else np.nan,
                'transect': 0,
            }
            rows.append(row)

            sample_offset += int(ping_cnt) * int(bytes_per_sample)

        return rows

    def _decode_position(self, x: float, y: float, nav_units=None):
        if nav_units == 3 and np.isfinite(y) and np.isfinite(x) and abs(y) <= 90 and abs(x) <= 180:
            lat = float(y)
            lon = float(x)
            epsg = self._convert_wgs_to_utm(lon, lat)
            self.humDat['epsg'] = f'EPSG:{epsg}'
            self.humDat['wgs'] = 'EPSG:4326'
            self.trans = pyproj.Proj(self.humDat['epsg'])
            e, n = self.trans(lon, lat)
            return lat, lon, float(e), float(n)

        if nav_units is None and np.isfinite(y) and np.isfinite(x) and abs(y) <= 90 and abs(x) <= 180:
            lat = float(y)
            lon = float(x)
            epsg = self._convert_wgs_to_utm(lon, lat)
            self.humDat['epsg'] = f'EPSG:{epsg}'
            self.humDat['wgs'] = 'EPSG:4326'
            self.trans = pyproj.Proj(self.humDat['epsg'])
            e, n = self.trans(lon, lat)
            return lat, lon, float(e), float(n)

        self.humDat['epsg'] = 'UNKNOWN'
        self.humDat['wgs'] = 'EPSG:4326'
        self.trans = lambda lon, lat: (lon, lat)
        if np.isfinite(x) and np.isfinite(y):
            return np.nan, np.nan, float(x), float(y)
        return np.nan, np.nan, np.nan, np.nan

    def _convert_wgs_to_utm(self, lon: float, lat: float):
        utm_band = str(int((np.floor((lon + 180) / 6) % 60) + 1))
        if len(utm_band) == 1:
            utm_band = '0' + utm_band
        if lat >= 0:
            return '326' + utm_band
        return '327' + utm_band

    def _map_beam(self, channel_number: int, type_of_channel, chan_idx: int):
        if type_of_channel == 1:
            return 2
        if type_of_channel == 2:
            return 3

        if chan_idx % 2 == 0:
            return 2
        return 3

    def _map_freq_band(self, channel_number: int, chan_idx: int = None):
        if channel_number in (0, 1):
            return 'low'
        if channel_number in (2, 3):
            return 'high'
        if channel_number in (4, 5):
            return 'vhigh'

        if chan_idx is not None:
            pair_idx = int(chan_idx) // 2
            if pair_idx == 0:
                return 'low'
            if pair_idx == 1:
                return 'high'
            if pair_idx == 2:
                return 'vhigh'

        return None

    def _doUnitConversion(self, df: pd.DataFrame):
        if 'inst_dep_m' not in df.columns:
            if 'dep_m' in df.columns:
                df['inst_dep_m'] = df['dep_m']
            else:
                df['inst_dep_m'] = np.nan

        if 'dep_m' not in df.columns:
            df['dep_m'] = df['inst_dep_m']

        if 'dep_m' not in df.columns:
            df['dep_m'] = np.nan
        if 'speed_ms' not in df.columns:
            df['speed_ms'] = np.nan
        if 'instr_heading' not in df.columns:
            df['instr_heading'] = np.nan

        if 'f' in df.columns:
            invalid_f = ~np.isfinite(df['f'])
            df.loc[invalid_f, 'f'] = np.nan
            df.loc[invalid_f, 'f_min'] = np.nan
            df.loc[invalid_f, 'f_max'] = np.nan

        df['tempC'] = np.float32(self.tempC * 10)
        return df

    def _calcTrkDistTS(self, df: pd.DataFrame):
        ts = df['time_s'].to_numpy(dtype=float)
        ss = df['speed_ms'].fillna(0).to_numpy(dtype=float)

        if len(ts) == 0:
            df['trk_dist'] = []
            return df

        ds = np.zeros((len(ts),), dtype=float)
        if len(ts) > 1:
            d = np.maximum(0, (ts[1:] - ts[:-1]) * ss[1:])
            ds[1:] = d
            ds = np.cumsum(ds)

        df['trk_dist'] = ds
        return df

    def _recalcRecordNum(self):
        df = self.header_dat.reset_index(drop=True)
        df['record_num'] = df.index
        self.header_dat = df
        return

    def _splitBeamsToCSV(self):
        self.beamMeta = beamMeta = {}
        df = self.header_dat

        group_cols = ['beam']
        if 'freq_band' in df.columns:
            group_cols.append('freq_band')

        for group_key, group in df.groupby(group_cols):
            if isinstance(group_key, tuple):
                beam, freq_band = group_key
            else:
                beam, freq_band = group_key, None
            meta = {}

            if beam in [2, 3] and 'pixM' in group.columns and len(group) > 0:
                self.pixM = group['pixM'].iloc[0]

            beam_name = f'B00{int(beam)}'
            base_name = self._getBeamName(beam_name)
            if freq_band:
                meta['beamName'] = f'{base_name}_{freq_band}'
            else:
                meta['beamName'] = base_name
            meta['sonFile'] = self.sonFile

            group = self._getChunkID(group.copy())

            out_csv = f'{beam_name}_{meta["beamName"]}_meta.csv'
            out_csv = os.path.join(self.metaDir, out_csv)
            group.to_csv(out_csv, index=False)

            meta['metaCSV'] = out_csv
            key = beam_name if not freq_band else f'{beam_name}_{freq_band}'
            beamMeta[key] = meta

        return

    def _getBeamName(self, beam: str):
        if beam == 'B000':
            return 'ds_lowfreq'
        if beam == 'B001':
            return 'ds_highfreq'
        if beam == 'B002':
            return 'ss_port'
        if beam == 'B003':
            return 'ss_star'
        if beam == 'B004':
            return 'ds_vhighfreq'
        return 'unknown'

    def _getChunkID(self, df: pd.DataFrame):
        df.reset_index(drop=True, inplace=True)
        df['chunk_id'] = int(-1)

        chunk = 0
        start_idx = 0
        end_idx = self.nchunk

        while start_idx < len(df):
            df.iloc[start_idx:end_idx, df.columns.get_loc('chunk_id')] = int(chunk)
            chunk += 1
            start_idx = end_idx
            end_idx += self.nchunk

        last_chunk = df[df['chunk_id'] == chunk]
        if len(last_chunk) <= self.nchunk / 2:
            df.loc[df['chunk_id'] == chunk, 'chunk_id'] = chunk - 1

        return df



import os
import struct

import numpy as np
import pandas as pd
import pyproj


class jsf(object):

    def __init__(self, inFile: str, nchunk: int = 0, exportUnknown: bool = False):
        self.humFile = None
        self.isOnix = 0
        self.sonFile = inFile
        self.nchunk = nchunk
        self.exportUnknown = exportUnknown

        self.file_header_size = 0
        self.msg_header_size = 16
        self.msg80_header_size = 240

        self.humDat = {}
        self.headBytes = self.msg_header_size + self.msg80_header_size
        self.son8bit = False
        self.sample_dtype = '<u2'

    def _getFileLen(self):
        self.file_len = os.path.getsize(self.sonFile)
        return

    def _parseFileHeader(self):
        self.file_header = {
            'format': 'JSF',
            'message_header_size': self.msg_header_size,
            'message80_header_size': self.msg80_header_size,
            'endianness': 'little',
        }

        out_file = os.path.join(self.metaDir, 'DAT_meta.csv')
        pd.DataFrame.from_dict(self.file_header, orient='index').T.to_csv(out_file, index=False)
        self.datMetaFile = out_file
        return

    def _parsePingHeader(self):
        rows = []

        with open(self.sonFile, 'rb') as file:
            file_len = self.file_len
            i = self.file_header_size

            while i + self.msg_header_size <= file_len:
                file.seek(i)
                msg_head = file.read(self.msg_header_size)
                if len(msg_head) < self.msg_header_size:
                    break

                marker = struct.unpack_from('<H', msg_head, 0)[0]
                if marker != 0x1601:
                    i += 1
                    continue

                protocol_version = msg_head[2]
                message_type = struct.unpack_from('<H', msg_head, 4)[0]
                subsystem_number = msg_head[7]
                channel = msg_head[8]
                msg_size = struct.unpack_from('<i', msg_head, 12)[0]

                if msg_size <= 0:
                    i += 1
                    continue

                next_i = i + self.msg_header_size + msg_size
                if next_i > file_len:
                    break

                if message_type == 80 and msg_size >= self.msg80_header_size:
                    file.seek(i + self.msg_header_size)
                    msg80 = file.read(self.msg80_header_size)
                    rows.append(self._decode_msg80(i, protocol_version, subsystem_number, channel, msg80, msg_size))

                i = next_i

        df = pd.DataFrame.from_dict(rows)

        if len(df) == 0:
            raise ValueError('No JSF Message Type 80 records were parsed.')

        df = self._doUnitConversion(df)
        df.sort_values(by=['time_s', 'beam'], inplace=True)
        df.reset_index(drop=True, inplace=True)
        df = self._calcTrkDistTS(df)
        df['record_num'] = np.arange(len(df), dtype=np.int64)

        out_test = os.path.join(self.metaDir, 'All-JSF-Sonar-MetaData.csv')
        df.to_csv(out_test, index=False)

        self.header_dat = df
        return

    def _decode_msg80(self, record_start, protocol_version, subsystem_number, channel, msg80, msg_size):
        time_since_1970 = struct.unpack_from('<i', msg80, 0)[0]
        ping_number = struct.unpack_from('<I', msg80, 8)[0]
        msb1 = struct.unpack_from('<H', msg80, 16)[0]
        lsb1 = struct.unpack_from('<H', msg80, 18)[0]
        lsb2 = struct.unpack_from('<H', msg80, 20)[0]
        validity_flag = struct.unpack_from('<H', msg80, 30)[0]
        data_format = struct.unpack_from('<h', msg80, 34)[0]

        longitude_raw = struct.unpack_from('<i', msg80, 80)[0]
        latitude_raw = struct.unpack_from('<i', msg80, 84)[0]
        coord_units = struct.unpack_from('<h', msg80, 88)[0]

        samples_lsb = struct.unpack_from('<H', msg80, 114)[0]
        samples_msb = (msb1 >> 8) & 0x0F
        samples = int(samples_lsb + (samples_msb << 16))

        sample_interval_ns = struct.unpack_from('<I', msg80, 116)[0]

        start_freq_dahz = struct.unpack_from('<H', msg80, 126)[0]
        end_freq_dahz = struct.unpack_from('<H', msg80, 128)[0]

        depth_mm = struct.unpack_from('<i', msg80, 136)[0]
        altitude_mm = struct.unpack_from('<i', msg80, 144)[0]
        sound_speed = struct.unpack_from('<f', msg80, 148)[0]

        weighting_factor = struct.unpack_from('<h', msg80, 168)[0]
        compass_heading = struct.unpack_from('<H', msg80, 172)[0]
        pitch_raw = struct.unpack_from('<h', msg80, 174)[0]
        roll_raw = struct.unpack_from('<h', msg80, 176)[0]

        course_tenths = struct.unpack_from('<h', msg80, 192)[0]
        speed_tenths_knots = struct.unpack_from('<h', msg80, 194)[0]

        milli_seconds_today = struct.unpack_from('<I', msg80, 200)[0]
        max_abs_adc_raw = struct.unpack_from('<H', msg80, 204)[0]
        water_temp_tenths = struct.unpack_from('<h', msg80, 226)[0]
        layback_m = struct.unpack_from('<f', msg80, 228)[0]
        cable_out_dm = struct.unpack_from('<H', msg80, 236)[0]

        ints_per_sample = 2 if data_format in [1, 9] else 1
        ping_cnt = int(samples * ints_per_sample)

        payload_bytes = max(int(msg_size) - self.msg80_header_size, 0)
        max_ints = payload_bytes // 2
        if max_ints >= 0 and ping_cnt > max_ints:
            ping_cnt = int(max_ints)

        son_offset = self.msg_header_size + self.msg80_header_size

        beam = self._map_beam(subsystem_number, channel)
        freq_band = self._map_freq_band(subsystem_number)

        lat, lon, e, n = self._decode_position(latitude_raw, longitude_raw, coord_units)

        heading = float(compass_heading) / 100.0
        pitch = float(pitch_raw) * 180.0 / 32768.0
        roll = float(roll_raw) * 180.0 / 32768.0
        course_frac = float((lsb1 >> 8) & 0xFF) / 100.0
        course = float(course_tenths) / 10.0 + course_frac

        speed_frac = float(lsb2 & 0x0F) / 100.0
        speed_kn = float(speed_tenths_knots) / 10.0 + speed_frac

        if sound_speed <= 0 or not np.isfinite(sound_speed):
            sound_speed = 1500.0

        pix_m = np.nan
        if sample_interval_ns > 0:
            pix_m = (float(sound_speed) * (sample_interval_ns / 1e9)) / 2.0

        start_freq_khz = start_freq_dahz / 100.0 if start_freq_dahz > 0 else np.nan
        end_freq_khz = end_freq_dahz / 100.0 if end_freq_dahz > 0 else np.nan
        if np.isfinite(start_freq_khz) and np.isfinite(end_freq_khz):
            f = (start_freq_khz + end_freq_khz) / 2.0
        elif np.isfinite(start_freq_khz):
            f = start_freq_khz
        elif np.isfinite(end_freq_khz):
            f = end_freq_khz
        else:
            f = np.nan

        dep_m = np.nan
        if depth_mm > 0 and (validity_flag & (1 << 9)):
            dep_m = depth_mm / 1000.0

        altitude_m = np.nan
        if altitude_mm > 0 and (validity_flag & (1 << 6)):
            altitude_m = altitude_mm / 1000.0

        if not np.isfinite(dep_m) and np.isfinite(altitude_m):
            dep_m = altitude_m

        time_s = float(time_since_1970)
        if milli_seconds_today > 0:
            frac = (milli_seconds_today % 1000) / 1000.0
            time_s = float(time_since_1970) + frac

        row = {
            'index': int(record_start),
            'son_offset': int(son_offset),
            'protocol_version': int(protocol_version),
            'message_type': 80,
            'subsystem_number': int(subsystem_number),
            'channel': int(channel),
            'beam': int(beam),
            'ping_number': int(ping_number),
            'time_s': float(time_s),
            'ping_cnt': int(ping_cnt),
            'data_format': int(data_format),
            'weighting_factor': int(weighting_factor),
            'max_abs_adc_raw': int(max_abs_adc_raw),
            'f': float(f) if np.isfinite(f) else np.nan,
            'f_min': float(start_freq_khz) if np.isfinite(start_freq_khz) else np.nan,
            'f_max': float(end_freq_khz) if np.isfinite(end_freq_khz) else np.nan,
            'freq_band': freq_band,
            'pixM': float(pix_m) if np.isfinite(pix_m) else np.nan,
            'instr_heading': float(heading),
            'pitch': float(pitch),
            'roll': float(roll),
            'yaw': np.nan,
            'course': float(course),
            'speed_ms': float(speed_kn) * 0.514444,
            'inst_dep_m': float(dep_m) if np.isfinite(dep_m) else np.nan,
            'dep_m': float(dep_m) if np.isfinite(dep_m) else np.nan,
            'altitude': float(altitude_m) if np.isfinite(altitude_m) else np.nan,
            'lat': lat,
            'lon': lon,
            'e': e,
            'n': n,
            'validity_flag': int(validity_flag),
            'water_temp_raw': int(water_temp_tenths),
            'layback_m': float(layback_m),
            'cable_out_m': float(cable_out_dm) / 10.0,
            'transect': 0,
        }

        return row

    def _decode_position(self, latitude_raw: int, longitude_raw: int, coord_units: int):
        if coord_units == 2:
            lat = latitude_raw / (10000.0 * 60.0)
            lon = longitude_raw / (10000.0 * 60.0)
            if np.isfinite(lat) and np.isfinite(lon) and abs(lat) <= 90 and abs(lon) <= 180:
                epsg = self._convert_wgs_to_utm(lon, lat)
                self.humDat['epsg'] = f'EPSG:{epsg}'
                self.humDat['wgs'] = 'EPSG:4326'
                self.trans = pyproj.Proj(self.humDat['epsg'])
                e, n = self.trans(lon, lat)
                return float(lat), float(lon), float(e), float(n)

        if coord_units == 1:
            x_m = longitude_raw / 1000.0
            y_m = latitude_raw / 1000.0
        elif coord_units == 3:
            x_m = longitude_raw / 10.0
            y_m = latitude_raw / 10.0
        elif coord_units == 4:
            x_m = longitude_raw / 100.0
            y_m = latitude_raw / 100.0
        else:
            x_m = np.nan
            y_m = np.nan

        self.humDat['epsg'] = 'UNKNOWN'
        self.humDat['wgs'] = 'EPSG:4326'
        self.trans = lambda lon, lat: (lon, lat)
        return np.nan, np.nan, float(x_m) if np.isfinite(x_m) else np.nan, float(y_m) if np.isfinite(y_m) else np.nan

    def _convert_wgs_to_utm(self, lon: float, lat: float):
        utm_band = str(int((np.floor((lon + 180) / 6) % 60) + 1))
        if len(utm_band) == 1:
            utm_band = '0' + utm_band
        if lat >= 0:
            return '326' + utm_band
        return '327' + utm_band

    def _map_beam(self, subsystem_number: int, channel: int):
        if channel == 0:
            return 2
        if channel == 1:
            return 3

        if subsystem_number == 0:
            return 1
        return 4

    def _map_freq_band(self, subsystem_number: int):
        if subsystem_number == 20:
            return 'low'
        if subsystem_number == 21:
            return 'high'
        if subsystem_number == 22:
            return 'vhigh'
        if subsystem_number == 70:
            return 'custom70'
        return None

    def _doUnitConversion(self, df: pd.DataFrame):
        if 'inst_dep_m' not in df.columns:
            if 'dep_m' in df.columns:
                df['inst_dep_m'] = df['dep_m']
            else:
                df['inst_dep_m'] = np.nan

        if 'dep_m' not in df.columns:
            df['dep_m'] = df['inst_dep_m']

        if 'altitude' in df.columns:
            alt = pd.to_numeric(df['altitude'], errors='coerce')
            inst = pd.to_numeric(df['inst_dep_m'], errors='coerce')
            dep = pd.to_numeric(df['dep_m'], errors='coerce')

            inst_fallback = (~np.isfinite(inst)) & np.isfinite(alt)
            dep_fallback = (~np.isfinite(dep)) & np.isfinite(alt)

            if inst_fallback.any():
                df.loc[inst_fallback, 'inst_dep_m'] = alt.loc[inst_fallback]
            if dep_fallback.any():
                df.loc[dep_fallback, 'dep_m'] = alt.loc[dep_fallback]

        df['tempC'] = np.float32(self.tempC * 10)

        if 'water_temp_raw' in df.columns:
            is_valid_temp = (df['validity_flag'] & (1 << 8)) > 0
            df.loc[is_valid_temp, 'tempC'] = df.loc[is_valid_temp, 'water_temp_raw'] / 10.0

        return df

    def _calcTrkDistTS(self, df: pd.DataFrame):
        ts = df['time_s'].to_numpy(dtype=float)
        ss = df['speed_ms'].fillna(0).to_numpy(dtype=float)

        if len(ts) == 0:
            df['trk_dist'] = []
            return df

        ds = np.zeros((len(ts),), dtype=float)
        if len(ts) > 1:
            d = np.maximum(0, (ts[1:] - ts[:-1]) * ss[1:])
            ds[1:] = d
            ds = np.cumsum(ds)

        df['trk_dist'] = ds
        return df

    def _recalcRecordNum(self):
        df = self.header_dat.reset_index(drop=True)
        df['record_num'] = df.index
        self.header_dat = df
        return

    def _splitBeamsToCSV(self):
        self.beamMeta = beamMeta = {}
        df = self.header_dat

        group_cols = ['beam']
        if 'freq_band' in df.columns:
            group_cols.append('freq_band')

        for group_key, group in df.groupby(group_cols):
            if isinstance(group_key, tuple):
                beam, freq_band = group_key
            else:
                beam, freq_band = group_key, None
            meta = {}

            if beam in [2, 3] and 'pixM' in group.columns and len(group) > 0:
                self.pixM = group['pixM'].iloc[0]

            beam_name = f'B00{int(beam)}'
            base_name = self._getBeamName(beam_name)
            if freq_band:
                meta['beamName'] = f'{base_name}_{freq_band}'
            else:
                meta['beamName'] = base_name
            meta['sonFile'] = self.sonFile

            group = self._getChunkID(group.copy())

            out_csv = f'{beam_name}_{meta["beamName"]}_meta.csv'
            out_csv = os.path.join(self.metaDir, out_csv)
            group.to_csv(out_csv, index=False)

            meta['metaCSV'] = out_csv
            key = beam_name if not freq_band else f'{beam_name}_{freq_band}'
            beamMeta[key] = meta

        return

    def _getBeamName(self, beam: str):
        if beam == 'B000':
            return 'ds_lowfreq'
        if beam == 'B001':
            return 'ds_highfreq'
        if beam == 'B002':
            return 'ss_port'
        if beam == 'B003':
            return 'ss_star'
        if beam == 'B004':
            return 'ds_vhighfreq'
        return 'unknown'

    def _getChunkID(self, df: pd.DataFrame):
        df.reset_index(drop=True, inplace=True)
        df['chunk_id'] = int(-1)

        chunk = 0
        start_idx = 0
        end_idx = self.nchunk

        while start_idx < len(df):
            df.iloc[start_idx:end_idx, df.columns.get_loc('chunk_id')] = int(chunk)
            chunk += 1
            start_idx = end_idx
            end_idx += self.nchunk

        last_chunk = df[df['chunk_id'] == chunk]
        if len(last_chunk) <= self.nchunk / 2:
            df.loc[df['chunk_id'] == chunk, 'chunk_id'] = chunk - 1

        return df



'''
Dependency of PINGMapper: https://github.com/CameronBodine/PINGMapper

Repository: https://github.com/CameronBodine/PINGVerter
PyPi: https://pypi.org/project/pingverter/ 

Developed by Cameron S. Bodine

###############
Acknowledgments
###############

None of this work would have been possible without the following repositories:

PyHum: https://github.com/dbuscombe-usgs/PyHum
SL3Reader: https://github.com/halmaia/SL3Reader
sonarlight: https://github.com/KennethTM/sonarlight


MIT License

Copyright (c) 2024 Cameron S. Bodine

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import pandas as pd
import numpy as np

class sonar(object):

    def __init__(self):

        print('sonar class poop')

        pass

    def _parsePingHeader(self, in_file: str, out_file: str=None):
        '''
        '''

        # Get file length
        file_len = self.file_len = os.path.getsize(in_file)

        # Initialize counter
        i = 0
        chunk_i = 0
        chunk = 0


        header_dat_all = []

        frame_offset = []

        chunk_id = []

        file = open(in_file, 'rb')

        # Decode ping header
        while i < file_len:

            print(i)

            # Get header data at offset i
            header_dat, cpos = self._getPingHeader(file, i)

            # Add frame offset
            frame_offset.append(i)

            header_dat_all.append(header_dat)

            chunk_id.append(chunk)

            # update counter with current position
            i = cpos

            if chunk_i == self.nchunk:
                chunk_i = 0
                chunk += 1
            else:
                chunk_i += 1

        header_dat_all = pd.DataFrame.from_dict(header_dat_all)

        # Add in the frame offset
        header_dat_all['index'] = frame_offset

        # Add in the son_offset (headBytes for Humminbird)
        header_dat_all['son_offset'] = self.headBytes

        # Add chunk id
        header_dat_all['chunk_id'] = chunk_id

        # Do unit conversions
        header_dat_all = self._doUnitConversion(header_dat_all)
        

        # Drop spacer and unknown columns
        for col in header_dat_all.columns:
            if 'SP' in col:
                header_dat_all.drop(col, axis=1, inplace=True)

            if not self.exportUnknown and 'unknown' in col:
                header_dat_all.drop(col, axis=1, inplace=True)

        # Drop head_start
        header_dat_all.drop('head_start', axis=1, inplace=True)
        header_dat_all.drop('head_end', axis=1, inplace=True)

        # Update last chunk if too small (for rectification)
        lastChunk = header_dat_all[header_dat_all['chunk_id'] == chunk]
        if len(lastChunk) <= self.nchunk/2:
            header_dat_all.loc[header_dat_all['chunk_id'] == chunk, 'chunk_id'] = chunk-1


        # Save to csv
        if out_file:
            header_dat_all.to_csv(out_file, index=False)
        else:
            self.header_dat = header_dat_all

        return
    
    def _getPingHeader(self, file, i: int):

        # Get necessary attributes
        head_struct = self.son_struct
        length = self.frame_header_size # Account for start and end header

        # Move to offset
        file.seek(i)

        # Get the data
        buffer = file.read(length)

        # Read the data
        header = np.frombuffer(buffer, dtype=head_struct)

        out_dict = {}
        for name, typ in header.dtype.fields.items():
            out_dict[name] = header[name][0].item()

        # Next ping header is from current position + ping_cnt
        next_ping = int(file.tell() + header[0][-2])

        return out_dict, next_ping
